#include "postgres.h"

#include <unistd.h>

#include "funcapi.h"
#include "miscadmin.h"
#include "pgstat.h"

#include "access/transam.h"
#include "access/xact.h"
#include "access/xlog.h"
#include "access/xlog_internal.h"
#include "access/xloginsert.h"
#include "access/xlogutils.h"
#include "access/xlogreader.h"
#include "access/xlogrecord.h"
#include "replication/walreceiver.h"
#include "storage/fd.h"
#include "storage/proc.h"
#include "storage/buf_internals.h"
#include "utils/pg_lsn.h"

#ifdef PG_MODULE_MAGIC
PG_MODULE_MAGIC;
#endif

extern void XLogPrefetch(XLogRecPtr lsn);

static int
XLogReadLocalPage(XLogReaderState *state, XLogRecPtr targetPagePtr,
				  int reqLen, XLogRecPtr targetRecPtr, char *cur_page,
				  TimeLineID *pageTLI);
static XLogRecPtr
XLogFindNextRecord(XLogReaderState *state, XLogRecPtr RecPtr);

#define WAIT_WAL_DELAY  1000 /* 1 msec */
#define BLOCK_HASH_SIZE 4001 /* Size of block LRU cache, may be it is better to use information about free RAM instead of hardcoded constant */
#define FILE_HASH_SIZE  64   /* Size of opened files hash */
#define STAT_REFRESH_PERIOD 1024 /* Refresh backend status rate */
#define PREFETCH_LEAD   0 /* Lead in WAL for prefetch */

/*
 * Block LRU hash table is used to keep information about most recently prefetched blocks.
 */
typedef struct BlockHashEntry
{
	struct BlockHashEntry* next;
	struct BlockHashEntry* prev;
	struct BlockHashEntry* collision;
	BufferTag tag;
	uint32 hash;
} BlockHashEntry;

static BlockHashEntry* block_hash_table[BLOCK_HASH_SIZE];
static size_t block_hash_size;
static BlockHashEntry lru = {&lru, &lru};
static TimeLineID replay_timeline;

/*
 * Yet another L2-list implementation
 */
static void
unlink_block(BlockHashEntry* entry)
{
	entry->next->prev = entry->prev;
	entry->prev->next = entry->next;
}

static void
link_block_after(BlockHashEntry* head, BlockHashEntry* entry)
{
	entry->next = head->next;
	entry->prev = head;
	head->next->prev = entry;
	head->next = entry;
}

/*
 * Put block in LRU hash or link it to the head of LRU list. Returns true if block was not present in hash, false otherwise.
 */
static bool
put_block_in_cache(BufferTag* tag)
{
	uint32 hash;
	BlockHashEntry* entry;

	hash = BufTableHashCode(tag) % BLOCK_HASH_SIZE;
	for (entry = block_hash_table[hash]; entry != NULL; entry = entry->collision)
	{
		if (BUFFERTAGS_EQUAL(entry->tag, *tag))
		{
			unlink_block(entry);
			link_block_after(&lru, entry);
			return false;
		}
	}
	if (block_hash_size == BLOCK_HASH_SIZE)
	{
		BlockHashEntry* victim = lru.prev;
		BlockHashEntry** epp = &block_hash_table[victim->hash];
		while (*epp != victim)
			epp = &(*epp)->collision;
		*epp = (*epp)->collision;
		unlink_block(victim);
		entry = victim;
	}
	else
	{
		entry = (BlockHashEntry*)palloc(sizeof(BlockHashEntry));
		block_hash_size += 1;
	}
	entry->tag = *tag;
	entry->hash = hash;
	entry->collision = block_hash_table[hash];
	block_hash_table[hash] = entry;
	link_block_after(&lru, entry);

	return true;
}

/*
 * Hash of of opened files. It seems to be simpler to maintain own cache rather than provide SMgrRelation for smgr functions.
 */
typedef struct FileHashEntry
{
    BufferTag tag;
	File file;
} FileHashEntry;

static FileHashEntry file_hash_table[FILE_HASH_SIZE];

static File
XLogOpenFile(BufferTag* tag)
{
	BufferTag segment_tag = *tag;
	uint32 hash;
	char* path;
	File file;

	/* Transform block number into segment number */
	segment_tag.blockNum /= RELSEG_SIZE;
	hash = BufTableHashCode(&segment_tag) % FILE_HASH_SIZE;

	if (BUFFERTAGS_EQUAL(file_hash_table[hash].tag, segment_tag))
		return file_hash_table[hash].file;

	path = relpathperm(tag->rnode, tag->forkNum);
	if (segment_tag.blockNum > 0)
	{
		char* fullpath = psprintf("%s.%d", path, segment_tag.blockNum);
		pfree(path);
		path = fullpath;
	}
	file = PathNameOpenFile(path, O_RDONLY | PG_BINARY);
	pfree(path);

	if (file >= 0)
	{
		if (file_hash_table[hash].tag.rnode.dbNode != 0)
			FileClose(file_hash_table[hash].file);

		file_hash_table[hash].file = file;
		file_hash_table[hash].tag = segment_tag;
	}
	return file;
}

/*
 * Our backend doesn't receive any notifications about WAL progress, so we have to use sleep
 * to wait until requested information is available
 */
static void
XLogWaitWAL(void)
{
	CHECK_FOR_INTERRUPTS();
	pg_usleep(WAIT_WAL_DELAY);
}

/*
 * Main function: perform prefetch of blocks referenced by WAL records starting from given LSN or from WAL replay position if lsn=0
 */
void
XLogPrefetch(XLogRecPtr lsn)
{
	XLogReaderState *xlogreader;
	long n_prefetched = 0;

	/* Dirty hack: prevent recovery conflict */
	MyPgXact->xmin = InvalidTransactionId;

	memset(file_hash_table, 0, sizeof file_hash_table);
	memset(block_hash_table, 0, sizeof block_hash_table);

	xlogreader = XLogReaderAllocate(wal_segment_size, &XLogReadLocalPage, NULL);

	if (!xlogreader)
		ereport(ERROR,
				(errcode(ERRCODE_OUT_OF_MEMORY),
				 errmsg("out of memory"),
				 errdetail("Failed while allocating a WAL reading processor.")));

	if (lsn == InvalidXLogRecPtr)
		lsn = GetXLogReplayRecPtr(NULL); /* By default start with replay LSN */

	while (true)
	{
		char *errormsg;
		int	block_id;
		XLogRecPtr replay_lsn = GetXLogReplayRecPtr(&replay_timeline);
		XLogRecord *record;

		/*
		 * If current position is behind current replay LSN, then move it forward: we do not want to perform useless job and prefetch
		 * blocks for already processed WAL records
		 */
		if (lsn != InvalidXLogRecPtr || replay_lsn >= xlogreader->EndRecPtr)
		{
			XLogRecPtr prefetch_lsn = XLogFindNextRecord(xlogreader, Max(lsn, replay_lsn) + PREFETCH_LEAD);
			if (prefetch_lsn == InvalidXLogRecPtr)
			{
				XLogWaitWAL();
				continue;
			}
			lsn = prefetch_lsn;
		}

		record = XLogReadRecord(xlogreader, lsn, &errormsg);

		if (record != NULL)
		{
			lsn = InvalidXLogRecPtr; /* continue with next record */

			/* Loop through blocks referenced by this WAL record */
			for (block_id = 0; block_id <= xlogreader->max_block_id; block_id++)
			{
				BufferTag tag;
				File      file;

				if (!XLogRecGetBlockTag(xlogreader, block_id, &tag.rnode, &tag.forkNum, &tag.blockNum))
					continue;

				/* Check if block already prefetched */
				if (!put_block_in_cache(&tag))
					continue;

				file = XLogOpenFile(&tag);
				if (file >= 0)
				{
					off_t offs = (off_t) BLCKSZ * (tag.blockNum % ((BlockNumber) RELSEG_SIZE));
					int rc = FilePrefetch(file, offs, BLCKSZ, WAIT_EVENT_DATA_FILE_PREFETCH);
					if (rc != 0)
						elog(ERROR, "Failed to prefetch file: %m");
					else if (++n_prefetched % STAT_REFRESH_PERIOD == 0)
					{
						char buf[1024];
						sprintf(buf, "Prefetch %ld blocks at LSN %lx, replay LSN %lx", n_prefetched, xlogreader->EndRecPtr, replay_lsn);
						pgstat_report_activity(STATE_RUNNING, buf);
						elog(DEBUG1, "%s", buf);
					}
				}
				else
					elog(LOG, "File segment doesn't exists");
			}
		}
		else
			XLogWaitWAL();
	}
}

PG_FUNCTION_INFO_V1(pg_wal_prefetch);
Datum pg_wal_prefetch(PG_FUNCTION_ARGS)
{
	XLogRecPtr lsn = PG_GETARG_LSN(0);

	XLogPrefetch(lsn);

	PG_RETURN_VOID();
}

/*
 * XLogRead function is reading data from WAL pages. It is used by XLogReadLocalPage.
 * I have to copy this function from xlogutils.c because it is declared as static.
 */
static void
XLogRead(char *buf, int segsize, TimeLineID tli, XLogRecPtr startptr,
		 Size count)
{
	char	   *p;
	XLogRecPtr	recptr;
	Size		nbytes;

	/* state maintained across calls */
	static int	sendFile = -1;
	static XLogSegNo sendSegNo = 0;
	static TimeLineID sendTLI = 0;
	static uint32 sendOff = 0;

	Assert(segsize == wal_segment_size);

	p = buf;
	recptr = startptr;
	nbytes = count;

	while (nbytes > 0)
	{
		uint32		startoff;
		int			segbytes;
		int			readbytes;

		startoff = XLogSegmentOffset(recptr, segsize);

		/* Do we need to switch to a different xlog segment? */
		if (sendFile < 0 || !XLByteInSeg(recptr, sendSegNo, segsize) ||
			sendTLI != tli)
		{
			char		path[MAXPGPATH];

			if (sendFile >= 0)
				close(sendFile);

			XLByteToSeg(recptr, sendSegNo, segsize);

			XLogFilePath(path, tli, sendSegNo, segsize);

			sendFile = BasicOpenFile(path, O_RDONLY | PG_BINARY);

			if (sendFile < 0)
			{
				if (errno == ENOENT)
					ereport(ERROR,
							(errcode_for_file_access(),
							 errmsg("requested WAL segment %s has already been removed",
									path)));
				else
					ereport(ERROR,
							(errcode_for_file_access(),
							 errmsg("could not open file \"%s\": %m",
									path)));
			}
			sendOff = 0;
			sendTLI = tli;
		}

		/* Need to seek in the file? */
		if (sendOff != startoff)
		{
			if (lseek(sendFile, (off_t) startoff, SEEK_SET) < 0)
			{
				char		path[MAXPGPATH];

				XLogFilePath(path, tli, sendSegNo, segsize);

				ereport(ERROR,
						(errcode_for_file_access(),
						 errmsg("could not seek in log segment %s to offset %u: %m",
								path, startoff)));
			}
			sendOff = startoff;
		}

		/* How many bytes are within this segment? */
		if (nbytes > (segsize - startoff))
			segbytes = segsize - startoff;
		else
			segbytes = nbytes;

		pgstat_report_wait_start(WAIT_EVENT_WAL_READ);
		readbytes = read(sendFile, p, segbytes);
		pgstat_report_wait_end();
		if (readbytes <= 0)
		{
			char		path[MAXPGPATH];

			XLogFilePath(path, tli, sendSegNo, segsize);

			ereport(ERROR,
					(errcode_for_file_access(),
					 errmsg("could not read from log segment %s, offset %u, length %lu: %m",
							path, sendOff, (unsigned long) segbytes)));
		}

		/* Update state for read */
		recptr += readbytes;

		sendOff += readbytes;
		nbytes -= readbytes;
		p += readbytes;
	}
}

/*
 * Almost copy of read_local_xlog_page from xlogutils.c, but it reads until flush position of WAL receiver, rather then replay position.
 */
static int
XLogReadLocalPage(XLogReaderState *state, XLogRecPtr targetPagePtr,
				  int reqLen, XLogRecPtr targetRecPtr, char *cur_page,
				  TimeLineID *pageTLI)
{
	XLogRecPtr	read_upto,
				loc;
	int			count;

	loc = targetPagePtr + reqLen;

	/* Loop waiting for xlog to be available if necessary */
	while (1)
	{
		read_upto =	WalRcv->receivedUpto;
		*pageTLI = replay_timeline;

		if (loc <= read_upto)
			break;

		XLogWaitWAL();
		CHECK_FOR_INTERRUPTS();
		pg_usleep(1000L);
	}

	if (targetPagePtr + XLOG_BLCKSZ <= read_upto)
	{
		/*
		 * more than one block available; read only that block, have caller
		 * come back if they need more.
		 */
		count = XLOG_BLCKSZ;
	}
	else if (targetPagePtr + reqLen > read_upto)
	{
		/* not enough data there */
		return -1;
	}
	else
	{
		/* enough bytes available to satisfy the request */
		count = read_upto - targetPagePtr;
	}

	/*
	 * Even though we just determined how much of the page can be validly read
	 * as 'count', read the whole page anyway. It's guaranteed to be
	 * zero-padded up to the page boundary if it's incomplete.
	 */
	XLogRead(cur_page, state->wal_segment_size, *pageTLI, targetPagePtr, XLOG_BLCKSZ);


	/* number of valid bytes in the buffer */
	return count;
}

/*
 * Function XLogFindNextRecord is available only for frontend, so has to copy it here together with static ReadPageInternal used by XLogFindNextRecord :(
 */
static int
ReadPageInternal(XLogReaderState *state, XLogRecPtr pageptr, int reqLen)
{
	int			readLen;
	uint32		targetPageOff;
	XLogSegNo	targetSegNo;
	XLogPageHeader hdr;

	Assert((pageptr % XLOG_BLCKSZ) == 0);

	XLByteToSeg(pageptr, targetSegNo, state->wal_segment_size);
	targetPageOff = XLogSegmentOffset(pageptr, state->wal_segment_size);

	/* check whether we have all the requested data already */
	if (targetSegNo == state->readSegNo && targetPageOff == state->readOff &&
		reqLen < state->readLen)
		return state->readLen;

	/*
	 * Data is not in our buffer.
	 *
	 * Every time we actually read the page, even if we looked at parts of it
	 * before, we need to do verification as the read_page callback might now
	 * be rereading data from a different source.
	 *
	 * Whenever switching to a new WAL segment, we read the first page of the
	 * file and validate its header, even if that's not where the target
	 * record is.  This is so that we can check the additional identification
	 * info that is present in the first page's "long" header.
	 */
	if (targetSegNo != state->readSegNo && targetPageOff != 0)
	{
		XLogRecPtr	targetSegmentPtr = pageptr - targetPageOff;

		readLen = state->read_page(state, targetSegmentPtr, XLOG_BLCKSZ,
								   state->currRecPtr,
								   state->readBuf, &state->readPageTLI);
		if (readLen < 0)
			goto err;

		/* we can be sure to have enough WAL available, we scrolled back */
		Assert(readLen == XLOG_BLCKSZ);

		if (!XLogReaderValidatePageHeader(state, targetSegmentPtr,
										  state->readBuf))
			goto err;
	}

	/*
	 * First, read the requested data length, but at least a short page header
	 * so that we can validate it.
	 */
	readLen = state->read_page(state, pageptr, Max(reqLen, SizeOfXLogShortPHD),
							   state->currRecPtr,
							   state->readBuf, &state->readPageTLI);
	if (readLen < 0)
		goto err;

	Assert(readLen <= XLOG_BLCKSZ);

	/* Do we have enough data to check the header length? */
	if (readLen <= SizeOfXLogShortPHD)
		goto err;

	Assert(readLen >= reqLen);

	hdr = (XLogPageHeader) state->readBuf;

	/* still not enough */
	if (readLen < XLogPageHeaderSize(hdr))
	{
		readLen = state->read_page(state, pageptr, XLogPageHeaderSize(hdr),
								   state->currRecPtr,
								   state->readBuf, &state->readPageTLI);
		if (readLen < 0)
			goto err;
	}

	/*
	 * Now that we know we have the full header, validate it.
	 */
	if (!XLogReaderValidatePageHeader(state, pageptr, (char *) hdr))
		goto err;

	/* update read state information */
	state->readSegNo = targetSegNo;
	state->readOff = targetPageOff;
	state->readLen = readLen;

	return readLen;

err:
	XLogReaderInvalReadState(state);
	return -1;
}

/*
 * Function XLogFindNextRecord is available only for frontend, so has to copy it here.
 */
static XLogRecPtr
XLogFindNextRecord(XLogReaderState *state, XLogRecPtr RecPtr)
{
	XLogReaderState saved_state = *state;
	XLogRecPtr	tmpRecPtr;
	XLogRecPtr	found = InvalidXLogRecPtr;
	XLogPageHeader header;
	char	   *errormsg;

	Assert(!XLogRecPtrIsInvalid(RecPtr));

	/*
	 * skip over potential continuation data, keeping in mind that it may span
	 * multiple pages
	 */
	tmpRecPtr = RecPtr;
	while (true)
	{
		XLogRecPtr	targetPagePtr;
		int			targetRecOff;
		uint32		pageHeaderSize;
		int			readLen;

		/*
		 * Compute targetRecOff. It should typically be equal or greater than
		 * short page-header since a valid record can't start anywhere before
		 * that, except when caller has explicitly specified the offset that
		 * falls somewhere there or when we are skipping multi-page
		 * continuation record. It doesn't matter though because
		 * ReadPageInternal() is prepared to handle that and will read at
		 * least short page-header worth of data
		 */
		targetRecOff = tmpRecPtr % XLOG_BLCKSZ;

		/* scroll back to page boundary */
		targetPagePtr = tmpRecPtr - targetRecOff;

		/* Read the page containing the record */
		readLen = ReadPageInternal(state, targetPagePtr, targetRecOff);
		if (readLen < 0)
			goto err;

		header = (XLogPageHeader) state->readBuf;

		pageHeaderSize = XLogPageHeaderSize(header);

		/* make sure we have enough data for the page header */
		readLen = ReadPageInternal(state, targetPagePtr, pageHeaderSize);
		if (readLen < 0)
			goto err;

		/* skip over potential continuation data */
		if (header->xlp_info & XLP_FIRST_IS_CONTRECORD)
		{
			/*
			 * If the length of the remaining continuation data is more than
			 * what can fit in this page, the continuation record crosses over
			 * this page. Read the next page and try again. xlp_rem_len in the
			 * next page header will contain the remaining length of the
			 * continuation data
			 *
			 * Note that record headers are MAXALIGN'ed
			 */
			if (MAXALIGN(header->xlp_rem_len) > (XLOG_BLCKSZ - pageHeaderSize))
				tmpRecPtr = targetPagePtr + XLOG_BLCKSZ;
			else
			{
				/*
				 * The previous continuation record ends in this page. Set
				 * tmpRecPtr to point to the first valid record
				 */
				tmpRecPtr = targetPagePtr + pageHeaderSize
					+ MAXALIGN(header->xlp_rem_len);
				break;
			}
		}
		else
		{
			tmpRecPtr = targetPagePtr + pageHeaderSize;
			break;
		}
	}

	/*
	 * we know now that tmpRecPtr is an address pointing to a valid XLogRecord
	 * because either we're at the first record after the beginning of a page
	 * or we just jumped over the remaining data of a continuation.
	 */
	while (XLogReadRecord(state, tmpRecPtr, &errormsg) != NULL)
	{
		/* continue after the record */
		tmpRecPtr = InvalidXLogRecPtr;

		/* past the record we've found, break out */
		if (RecPtr <= state->ReadRecPtr)
		{
			found = state->ReadRecPtr;
			goto out;
		}
	}

err:
out:
	/* Reset state to what we had before finding the record */
	state->ReadRecPtr = saved_state.ReadRecPtr;
	state->EndRecPtr = saved_state.EndRecPtr;
	XLogReaderInvalReadState(state);

	return found;
}
